#!/usr/bin/env python3

# Copyright © 2020-2025, Meheret Tesfaye Batu <meherett.batu@gmail.com>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit


class CoinTypes:  # https://github.com/satoshilabs/slips/blob/master/slip-0044.md

    Adcoin: int = 161
    AkashNetwork: int = 118  # Cosmos
    Algorand: int = 283
    Anon: int = 220
    Aptos: int = 637
    Arbitrum: int = 60  # Ethereum
    Argoneum: int = 421
    Artax: int = 219
    Aryacoin: int = 357
    Asiacoin: int = 51
    Auroracoin: int = 85
    Avalanche: int = 9000
    Avian: int = 921
    Axe: int = 4242
    Axelar: int = 118  # Cosmos
    BandProtocol: int = 494
    Base: int = 60  # Ethereum
    Bata: int = 89
    BeetleCoin: int = 800
    BelaCoin: int = 73
    Binance: int = 714
    BitCloud: int = 218
    Bitcoin: int = 0
    BitcoinAtom: int = 185
    BitcoinCash: int = 145
    BitcoinCashSLP: int = 145  # Bitcoin-Cash
    BitcoinGold: int = 156
    BitcoinGreen: int = 222
    BitcoinPlus: int = 65
    BitcoinPrivate: int = 183
    BitcoinSV: int = 236
    BitcoinZ: int = 177
    Bitcore: int = 160
    BitSend: int = 91
    Blackcoin: int = 10
    Blocknode: int = 2941
    BlockStamp: int = 254
    Bolivarcoin: int = 278
    BritCoin: int = 70
    CanadaECoin: int = 34
    Cannacoin: int = 19
    Cardano: int = 1815
    Celo: int = 52752
    Chihuahua: int = 118  # Cosmos
    Clams: int = 23
    ClubCoin: int = 79
    Compcoin: int = 71
    Cosmos: int = 118
    CPUChain: int = 363
    CranePay: int = 2304
    Crave: int = 186
    Dash: int = 5
    DeepOnion: int = 305
    Defcoin: int = 1337
    Denarius: int = 116
    Diamond: int = 152
    DigiByte: int = 20
    Digitalcoin: int = 18
    Divi: int = 301
    Dogecoin: int = 3
    dYdX: int = 22000118
    eCash: int = 145  # Bitcoin-Cash
    ECoin: int = 115
    EDRCoin: int = 56
    eGulden: int = 78
    Einsteinium: int = 41
    Elastos: int = 2305
    Energi: int = 9797
    EOS: int = 194
    Ergo: int = 429
    Ethereum: int = 60
    EuropeCoin: int = 151
    Evrmore: int = 175
    ExclusiveCoin: int = 190
    Fantom: int = 60  # Ethereum
    Feathercoin: int = 8
    FetchAI: int = 118  # Cosmos
    Filecoin: int = 461
    Firo: int = 136  # ZCoin
    Firstcoin: int = 167
    FIX: int = 336
    Flashcoin: int = 120
    Flux: int = 19167
    Foxdcoin: int = 175
    FujiCoin: int = 75
    GameCredits: int = 101
    GCRCoin: int = 49
    GoByte: int = 176
    Gridcoin: int = 84
    GroestlCoin: int = 17
    Gulden: int = 87
    Harmony: int = 1023
    Helleniccoin: int = 168
    Hempcoin: int = 113
    Horizen: int = 121
    HuobiToken: int = 553
    Hush: int = 197
    Icon: int = 74
    Injective: int = 60  # Ethereum
    InsaneCoin: int = 68
    InternetOfPeople: int = 66
    IRISnet: int = 566
    IXCoin: int = 86
    Jumbucks: int = 26
    Kava: int = 459
    Kobocoin: int = 196
    Komodo: int = 141
    Landcoin: int = 63
    LBRYCredits: int = 140
    Linx: int = 114
    Litecoin: int = 2
    LitecoinCash: int = 192
    LitecoinZ: int = 221
    Lkrcoin: int = 557
    Lynx: int = 191
    Mazacoin: int = 13
    Megacoin: int = 217
    Metis: int = 60  # Ethereum
    Minexcoin: int = 182
    Monacoin: int = 22
    Monero: int = 128
    Monk: int = 214
    MultiversX: int = 508
    Myriadcoin: int = 90
    Namecoin: int = 7
    Nano: int = 165
    Navcoin: int = 130
    Near: int = 397
    Neblio: int = 146
    Neo: int = 888
    Neoscoin: int = 25
    Neurocoin: int = 110
    Neutron: int = 118  # Cosmos
    NewYorkCoin: int = 179
    NineChronicles: int = 567
    NIX: int = 400
    Novacoin: int = 50
    NuBits: int = 12
    NuShares: int = 11
    OKCash: int = 69
    OKTChain: int = 996
    Omni: int = 200
    Onix: int = 174
    Ontology: int = 1024
    Optimism: int = 60  # Ethereum
    Osmosis: int = 10000118
    Particl: int = 44
    Peercoin: int = 6
    Pesobit: int = 62
    Phore: int = 444
    PiNetwork: int = 314159
    Pinkcoin: int = 117
    Pivx: int = 119
    Polygon: int = 60  # Ethereum
    PoSWCoin: int = 47
    Potcoin: int = 81
    ProjectCoin: int = 533
    Putincoin: int = 122
    Qtum: int = 2301
    Rapids: int = 320
    Ravencoin: int = 175
    Reddcoin: int = 4
    Ripple: int = 144
    Ritocoin: int = 19169
    RSK: int = 137
    Rubycoin: int = 16
    Safecoin: int = 19165
    Saluscoin: int = 572
    Scribe: int = 545
    Secret: int = 529
    ShadowCash: int = 35
    Shentu: int = 118  # Cosmos
    Slimcoin: int = 63
    Smileycoin: int = 59
    Solana: int = 501
    Solarcoin: int = 58
    Stafi: int = 907
    Stash: int = 49344
    Stellar: int = 148
    Stratis: int = 105
    Sugarchain: int = 408
    Sui: int = 784
    Syscoin: int = 57
    Terra: int = 330
    Tezos: int = 1729
    Theta: int = 500
    ThoughtAI: int = 502
    TOACoin: int = 159
    Tron: int = 195
    TWINS: int = 970
    UltimateSecureCash: int = 112
    Unobtanium: int = 92
    Vcash: int = 127
    VeChain: int = 818
    Verge: int = 77
    Vertcoin: int = 28
    Viacoin: int = 14
    Vivo: int = 166
    Voxels: int = 129
    VPNCoin: int = 33
    Wagerr: int = 0  # Unknown (default to Bitcoin)
    Whitecoin: int = 559
    Wincoin: int = 181
    XinFin: int = 550
    XUEZ: int = 225
    Ycash: int = 347
    Zcash: int = 133
    ZClassic: int = 147
    Zetacoin: int = 719
    Zilliqa: int = 313
    ZooBC: int = 883
