#!/usr/bin/env python3

# Copyright © 2020-2024, Meheret Tesfaye Batu <meherett.batu@gmail.com>
#             2024, Eyoel Tadesse <eyoel_tadesse@proton.me>
# Distributed under the MIT software license, see the accompanying
# file COPYING or https://opensource.org/license/mit

from hdwallet.symbols import *


def test_symbols():

	assert ACC == "ACC"
	assert AKT == "AKT"
	assert ALGO == "ALGO"
	assert ANON == "ANON"
	assert APT == "APT"
	assert ARB == "ARB"
	assert AGM == "AGM"
	assert XAX == "XAX"
	assert AYA == "AYA"
	assert AC == "AC"
	assert AUR == "AUR"
	assert AVAX == "AVAX"
	assert AVN == "AVN"
	assert AXE == "AXE"
	assert AXL == "AXL"
	assert BAND == "BAND"
	assert BTA == "BTA"
	assert BEET == "BEET"
	assert BELA == "BELA"
	assert BNB == "BNB"
	assert BTDX == "BTDX"
	assert BTC == "BTC"
	assert BCA == "BCA"
	assert BCH == "BCH"
	assert SLP == "SLP"
	assert BTG == "BTG"
	assert BITG == "BITG"
	assert XBC == "XBC"
	assert BTCP == "BTCP"
	assert BSV == "BSV"
	assert BTCZ == "BTCZ"
	assert BTX == "BTX"
	assert BSD == "BSD"
	assert BLK == "BLK"
	assert BND == "BND"
	assert BST == "BST"
	assert BOLI == "BOLI"
	assert BRIT == "BRIT"
	assert CDN == "CDN"
	assert CCN == "CCN"
	assert ADA == "ADA"
	assert CELO == "CELO"
	assert HUA == "HUA"
	assert CLAM == "CLAM"
	assert CLUB == "CLUB"
	assert CMP == "CMP"
	assert ATOM == "ATOM"
	assert CPU == "CPU"
	assert CRP == "CRP"
	assert CRAVE == "CRAVE"
	assert DASH == "DASH"
	assert ONION == "ONION"
	assert DFC == "DFC"
	assert DNR == "DNR"
	assert DMD == "DMD"
	assert DGB == "DGB"
	assert DGC == "DGC"
	assert DIVI == "DIVI"
	assert DOGE == "DOGE"
	assert DYDX == "DYDX"
	assert XEC == "XEC"
	assert ECN == "ECN"
	assert EDRC == "EDRC"
	assert EFL == "EFL"
	assert EMC2 == "EMC2"
	assert ELA == "ELA"
	assert NRG == "NRG"
	assert EOS == "EOS"
	assert ERG == "ERG"
	assert ETH == "ETH"
	assert ERC == "ERC"
	assert EVR == "EVR"
	assert EXCL == "EXCL"
	assert FTM == "FTM"
	assert FTC == "FTC"
	assert FET == "FET"
	assert FIL == "FIL"
	assert FIRO == "FIRO"
	assert FRST == "FRST"
	assert FIX == "FIX"
	assert FLASH == "FLASH"
	assert FLUX == "FLUX"
	assert FOXD == "FOXD"
	assert FJC == "FJC"
	assert GAME == "GAME"
	assert GCR == "GCR"
	assert GBX == "GBX"
	assert GRC == "GRC"
	assert GRS == "GRS"
	assert NLG == "NLG"
	assert ONE == "ONE"
	assert HNC == "HNC"
	assert THC == "THC"
	assert ZEN == "ZEN"
	assert HT == "HT"
	assert HUSH == "HUSH"
	assert ICX == "ICX"
	assert INJ == "INJ"
	assert INSN == "INSN"
	assert IOP == "IOP"
	assert IRIS == "IRIS"
	assert IXC == "IXC"
	assert JBS == "JBS"
	assert KAVA == "KAVA"
	assert KOBO == "KOBO"
	assert KMD == "KMD"
	assert LDCN == "LDCN"
	assert LBC == "LBC"
	assert LINX == "LINX"
	assert LTC == "LTC"
	assert LCC == "LCC"
	assert LTZ == "LTZ"
	assert LKR == "LKR"
	assert LYNX == "LYNX"
	assert MZC == "MZC"
	assert MEC == "MEC"
	assert METIS == "METIS"
	assert MNX == "MNX"
	assert MONA == "MONA"
	assert XMR == "XMR"
	assert MONK == "MONK"
	assert EGLD == "EGLD"
	assert XMY == "XMY"
	assert NMC == "NMC"
	assert XNO == "XNO"
	assert NAV == "NAV"
	assert NEAR == "NEAR"
	assert NEBL == "NEBL"
	assert NEO == "NEO"
	assert NEOS == "NEOS"
	assert NRO == "NRO"
	assert NTRN == "NTRN"
	assert NYC == "NYC"
	assert NCG == "NCG"
	assert NIX == "NIX"
	assert NVC == "NVC"
	assert NBT == "NBT"
	assert NSR == "NSR"
	assert OK == "OK"
	assert OKT == "OKT"
	assert OMNI == "OMNI"
	assert ONX == "ONX"
	assert ONT == "ONT"
	assert OP == "OP"
	assert OSMO == "OSMO"
	assert PART == "PART"
	assert PPC == "PPC"
	assert PSB == "PSB"
	assert PHR == "PHR"
	assert PI == "PI"
	assert PINK == "PINK"
	assert PIVX == "PIVX"
	assert MATIC == "MATIC"
	assert POSW == "POSW"
	assert POT == "POT"
	assert PRJ == "PRJ"
	assert PUT == "PUT"
	assert QTUM == "QTUM"
	assert RPD == "RPD"
	assert RVN == "RVN"
	assert RDD == "RDD"
	assert XRP == "XRP"
	assert RITO == "RITO"
	assert RBTC == "RBTC"
	assert RBY == "RBY"
	assert SAFE == "SAFE"
	assert SLS == "SLS"
	assert SCRIBE == "SCRIBE"
	assert SCRT == "SCRT"
	assert SDC == "SDC"
	assert CTK == "CTK"
	assert SLM == "SLM"
	assert SMLY == "SMLY"
	assert SOL == "SOL"
	assert SLR == "SLR"
	assert FIS == "FIS"
	assert STASH == "STASH"
	assert XLM == "XLM"
	assert STRAT == "STRAT"
	assert SUGAR == "SUGAR"
	assert SUI == "SUI"
	assert SYS == "SYS"
	assert LUNA == "LUNA"
	assert XTZ == "XTZ"
	assert THETA == "THETA"
	assert THT == "THT"
	assert TOA == "TOA"
	assert TRX == "TRX"
	assert TWINS == "TWINS"
	assert USC == "USC"
	assert UNO == "UNO"
	assert VC == "VC"
	assert VET == "VET"
	assert XVG == "XVG"
	assert VTC == "VTC"
	assert VIA == "VIA"
	assert VIVO == "VIVO"
	assert VOX == "VOX"
	assert VASH == "VASH"
	assert WGR == "WGR"
	assert XWC == "XWC"
	assert WC == "WC"
	assert XDC == "XDC"
	assert XUEZ == "XUEZ"
	assert YEC == "YEC"
	assert ZEC == "ZEC"
	assert ZCL == "ZCL"
	assert ZET == "ZET"
	assert ZIL == "ZIL"
	assert ZBC == "ZBC"
